import re
import datetime
from pathlib import Path
from typing import Iterator
from abra_api.DataItem import DataItem
from abra_api.DataInfo import DataInfo
from abra_api.enums.DataItemTypeEnum import DataItemTypeEnum


class DataLoader:
    def __init__(self, directory: Path, file_encoding: str = 'UTF-8'):
        self.file_encoding = file_encoding
        if not directory.is_dir():
            raise ValueError('Directory {} not found!'.format(directory.absolute()))

        self.directory = directory

        """Regex for format 200347-0;2021/PD/sdgfsdfsdfsd/SML;2H3H000101;0"""
        self.line_regex = re.compile(r'^(\S+?);(\S+)/(.+?)(?:/\S+|);(\S+?);(\d)$')

    def parse_line(self, line: str) -> DataItem:
        """Parse single line into DataItem"""
        match = self.line_regex.search(line)
        if not match:
            raise ValueError('Line "{}" does not match data regex'.format(line))

        groups = match.groups()
        if len(groups) != 5:
            raise ValueError('Regex parsed incorrect number of groups for line: "{}"'.format(line))

        order_number, order_name_identifier, order_name, internal_id, state = groups

        state_to_data_type = {
            '0': DataItemTypeEnum.NEW,
            '1': DataItemTypeEnum.DONE
        }

        data_type = state_to_data_type.get(state)
        if not data_type:
            raise ValueError('Unknown state value {}'.format(state))

        return DataItem(order_number, order_name_identifier, order_name, internal_id, data_type)

    def parse_file(self, file: Path) -> DataInfo:
        """Parse file, returns DataItem"""
        if not file.is_file():
            raise ValueError('{} is not a file!'.format(file.absolute()))

        with file.open('r', encoding=self.file_encoding) as file_handle:
            lines = file_handle.readlines()
            if len(lines) > 1:
                raise ValueError('Loaded file {} contains more then one row!'.format(file.absolute()))

            modified_time = datetime.datetime.fromtimestamp(file.stat().st_mtime)
            created_time = datetime.datetime.fromtimestamp(file.stat().st_ctime)
            return DataInfo(file, created_time, modified_time, self.parse_line(lines[0]))

    def load_files(self) -> Iterator[DataInfo]:
        """Loads files and returns parsed data"""
        for nov_file in self.directory.glob('*.NOV'):
            yield self.parse_file(nov_file)

        for rea_file in self.directory.glob('*.REA'):
            yield self.parse_file(rea_file)
