import os
import logging
import hashlib
import tempfile
from filelock import FileLock
from importlib import import_module
from pathlib import Path, PurePath
from abra_api.DataLoader import DataLoader
from abra_api.EmailSender import EmailSender
from abra_api.config import Production


class DataSyncer:
    activated_filters = []

    def __init__(self, config: Production):
        self.config = config
        if config.BACKUP_DIRECTORY:
            backup_directory = Path(config.BACKUP_DIRECTORY)
            if not backup_directory.exists():
                raise Exception('Backup directory {} does not exists!'.format(backup_directory.absolute()))

            if not backup_directory.is_dir():
                raise Exception('Backup directory {} is not a directory!'.format(backup_directory.absolute()))
        else:
            backup_directory = None
        self.backup_directory = backup_directory

        # Activate filters
        for data_filter_cls_name in config.ENABLED_FILTERS:
            cls = data_filter_cls_name.split('.')[-1]
            self.activated_filters.append(getattr(import_module(data_filter_cls_name), cls)(config.__dict__))

        # Activate data loader
        self.data_loader = DataLoader(Path(config.INPUT_DIRECTORY), 'windows-1250')

    def close(self):
        # Cleanup
        # Close all activated filters
        for activated_filter in self.activated_filters:
            activated_filter.close()

    def sync(self, file_path: Path = None):
        lock_file_name = 'abra_api.{}'.format(hashlib.md5(file_path.absolute()).hexdigest()) if file_path else 'abra_api'
        file_lock_path = os.path.join(tempfile.gettempdir(), '{}.lock'.format(lock_file_name))

        # Lock this sync to prevent race conditions
        with FileLock(file_lock_path):
            report_log_error = []
            report_log_ok = []

            if file_path:
                found_files = [self.data_loader.parse_file(file_path)]
            else:
                found_files = self.data_loader.load_files()

            for file_info in found_files:
                is_handled = False
                for activated_filter in self.activated_filters:
                    if not activated_filter.is_handled(file_info):
                        continue

                    try:
                        activated_filter.process(file_info)
                        is_handled = True
                    except Exception as e:
                        logging.error('Failed to process file {}: {}'.format(file_info, e))
                        report_log_error.append('Failed to process file {}: {}.'.format(file_info.path.absolute(), e))

                # When file matched and successfully processed some filter, we assume it as processed
                # Move into BACKUP_DIRECTORY or just delete if no backup is configured
                if is_handled:
                    if self.backup_directory:
                        file_info.path.rename(PurePath(self.backup_directory, file_info.path.name))
                        report_log_ok.append('File {} has been processed and backed up successfully'.format(file_info.path.absolute()))
                    else:
                        file_info.path.unlink(missing_ok=True)
                        report_log_ok.append('File {} has been processed and deleted successfully'.format(file_info.path.absolute()))
                else:
                    logging.warning('File {} has matched no filter'.format(file_info))
                    report_log_error.append('File {} has matched no filter.'.format(file_info.path.absolute()))

            # Send report
            if self.config.REPORT_EMAIL and (len(report_log_error) > 0 or len(report_log_ok) > 0):
                email_sender = EmailSender(self.config.MAIL_SERVER, self.config.MAIL_PORT, self.config.MAIL_USE_SSL, self.config.MAIL_USE_TLS, self.config.MAIL_DEFAULT_SENDER)
                email_sender.login(self.config.MAIL_USERNAME, self.config.MAIL_PASSWORD)

                report_text_lines = [
                                        '===== ERRORS ====='
                                    ] + report_log_error + ['==================', '------Sync log------'] + report_log_ok + ['--------------------']

                email_sender.send(self.config.REPORT_EMAIL, 'AbraApi Sync report', '\n'.join(report_text_lines))