import imaplib
from types import TracebackType
from typing import Optional, Type, List, Iterator
from abra_api.enums.ImapStatusEnum import ImapStatusEnum
from abra_api import imap_utf7


class LogInContext:
    def __init__(self, imap: imaplib.IMAP4, user: str, password: str):
        self.imap = imap
        self.imap.login(user, password)

    def close(self) -> None:
        self.imap.logout()

    def __enter__(self) -> 'LogInContext':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self.close()


class ImapFolder:
    imap_folder_separator = b'.'

    def __init__(self, path_parts: List[str], tags: Optional[List[str]] = None):
        self.path_parts = path_parts
        self.tags = tags

    @staticmethod
    def from_str_list(directory_path_parts: List[str]) -> 'ImapFolder':
        return ImapFolder([directory_path_part for directory_path_part in directory_path_parts])

    @staticmethod
    def from_imap_format(folder: bytes) -> 'ImapFolder':
        tags_string, name_string = folder.split(b'"."')
        return ImapFolder(
            [imap_utf7.decode(item.strip()) for item in name_string.strip().strip(b'"').split(ImapFolder.imap_folder_separator)],
            [imap_utf7.decode(item.strip()) for item in tags_string.strip().strip(b'()').split(b'\\') if item.strip()]
        )

    def to_imap_format(self) -> bytes:
        return b'"' + self.imap_folder_separator.join([imap_utf7.encode(path_part) for path_part in self.path_parts]) + b'"'

    def to_path(self) -> str:
        return '/'.join(self.path_parts)

    def to_dict(self) -> dict:
        return {
            'path_parts': self.path_parts,
            'tags': self.tags,
        }

    def __eq__(self, other):
        if not isinstance(other, ImapFolder):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.to_path() == other.to_path()

    def __repr__(self):
        return '{}'.format(self.to_dict())


class ImapConnector:
    def __init__(self, host: str, use_ssl: bool = True):
        if use_ssl:
            self.imap = imaplib.IMAP4_SSL(host)
        else:
            self.imap = imaplib.IMAP4(host)

    def login(self, user: str, password: str):
        return LogInContext(self.imap, user, password)

    def get_folders(self) -> Iterator[ImapFolder]:
        status, result = self.imap.list()
        if ImapStatusEnum(status) != ImapStatusEnum.OK:
            raise Exception('Failed to list IMAP folders: {}'.format(result))

        for folder in result:
            yield ImapFolder.from_imap_format(folder)

    def create_folder(self, folder: ImapFolder):
        status, result = self.imap.create(folder.to_imap_format())
        if ImapStatusEnum(status) != ImapStatusEnum.OK:
            raise Exception('Failed to create IMAP folders: {}'.format(result))

    def subscribe_to_folder(self, folder: ImapFolder):
        status, result = self.imap.subscribe(folder.to_imap_format())
        if ImapStatusEnum(status) != ImapStatusEnum.OK:
            raise Exception('Failed to subscribe IMAP folder: {}'.format(result))

    def unsubscribe_from_folder(self, folder: ImapFolder):
        status, result = self.imap.unsubscribe(folder.to_imap_format())
        if ImapStatusEnum(status) != ImapStatusEnum.OK:
            raise Exception('Failed to unsubscribe IMAP folder: {}'.format(result))

    def rename_folder(self, old_folder: ImapFolder, new_folder: ImapFolder):
        self.imap.rename(old_folder.to_imap_format(), new_folder.to_imap_format())

    def close(self):
        self.imap.__exit__()

    def __enter__(self) -> 'ImapConnector':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self.close()
