#!/usr/bin/python
"""Main entry-point into the 'abra_api' CLI application.

ABRA API is simple script to sync weird file export into IMAP server.

License: PROPRIETARY
Website: https://gitlab.salamek.cz/milan.kozak/abra-api

Command details:
    sync             Sync now.
    watch            Watch configured directory.
    test_report_mail Test report email config connection.
    dump             Dump loaded data into console.

Usage:
    abra-api sync [--config_prod] [-l DIR]
    abra-api watch (pool|inotify) [-p TIME] [-l DIR] [--config_prod]
    abra-api dump [--config_prod]
    abra-api test_report_mail [--config_prod]
    abra-api (-h | --help)


Options:
     --config_prod              Load the production configuration instead of dev
     -p TIME --pool_time=TIME   Time to wait between pools
     -l DIR --log_dir=DIR       Directory to log into
"""

import logging
import logging.handlers
import os.path
import time
from importlib import import_module
import sys
import signal
import yaml
import pyinotify
from functools import wraps
from docopt import docopt
from pathlib import Path
import abra_api as app_root
from abra_api.ImapConnector import ImapConnector
from abra_api.DataLoader import DataLoader
from abra_api.EmailSender import EmailSender
from abra_api.DataSyncer import DataSyncer

OPTIONS = docopt(__doc__)
APP_ROOT_FOLDER = os.path.abspath(os.path.dirname(app_root.__file__))


class CustomFormatter(logging.Formatter):
    LEVEL_MAP = {logging.FATAL: 'F', logging.ERROR: 'E', logging.WARN: 'W', logging.INFO: 'I', logging.DEBUG: 'D'}

    def format(self, record):
        record.levelletter = self.LEVEL_MAP[record.levelno]
        return super(CustomFormatter, self).format(record)


def setup_logging(name: str=None, level: int=logging.DEBUG) -> None:
    """Setup Google-Style logging for the entire application.
    At first I hated this but I had to use it for work, and now I prefer it. Who knew?
    From: https://github.com/twitter/commons/blob/master/src/python/twitter/common/log/formatters/glog.py
    Always logs DEBUG statements somewhere.
    Positional arguments:
    name -- Append this string to the log file filename.
    """
    log_to_disk = False
    if OPTIONS['--log_dir']:
        if not os.path.isdir(OPTIONS['--log_dir']):
            print('ERROR: Directory {} does not exist.'.format(OPTIONS['--log_dir']))
            sys.exit(1)
        if not os.access(OPTIONS['--log_dir'], os.W_OK):
            print('ERROR: No permissions to write to directory {}.'.format(OPTIONS['--log_dir']))
            sys.exit(1)
        log_to_disk = True

    fmt = '%(levelletter)s%(asctime)s.%(msecs).03d %(process)d %(filename)s:%(lineno)d] %(message)s'
    datefmt = '%m%d %H:%M:%S'
    formatter = CustomFormatter(fmt, datefmt)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.ERROR if log_to_disk else logging.DEBUG)
    console_handler.setFormatter(formatter)

    root = logging.getLogger()
    root.setLevel(level)
    root.addHandler(console_handler)

    if log_to_disk:
        file_name = os.path.join(OPTIONS['--log_dir'], 'abra_api_{}.log'.format(name))
        file_handler = logging.handlers.TimedRotatingFileHandler(file_name, when='d', backupCount=7)
        file_handler.setFormatter(formatter)
        root.addHandler(file_handler)


def get_config(config_class_string: str, yaml_files=None):
    """Load the Flask config from a class.
    Positional arguments:
    config_class_string -- string representation of a configuration class that will be loaded (e.g.
        'abra_api.config.Production').
    yaml_files -- List of YAML files to load. This is for testing, leave None in dev/production.
    Returns:
    A class object to be fed into app.config.from_object().
    """
    config_module, config_class = config_class_string.rsplit('.', 1)
    config_obj = getattr(import_module(config_module), config_class)

    # Load additional configuration settings.
    yaml_files = yaml_files or [f for f in [
        os.path.join('/', 'etc', 'abra-api', 'config.yml'),
        os.path.abspath(os.path.join(APP_ROOT_FOLDER, '..', 'config.yml')),
        os.path.join(APP_ROOT_FOLDER, 'config.yml'),
    ] if os.path.exists(f)]
    additional_dict = dict()
    for y in yaml_files:
        with open(y) as f:
            config_content = f.read()
            try:
                loaded_data = yaml.load(config_content, Loader=yaml.FullLoader)
            except AttributeError:
                # Handle older versions of yaml library
                loaded_data = yaml.load(config_content)
            if isinstance(loaded_data, dict):
                additional_dict.update(loaded_data)
            else:
                raise Exception('Failed to parse configuration {}'.format(y))

    # Merge the rest into the Flask app config.
    for key, value in additional_dict.items():
        setattr(config_obj, key, value)

    return config_obj


def parse_config():
    """Parses command line options for Flask.
    Returns:
    Config instance to pass into create_app().
    """
    # Figure out which class will be imported.
    if OPTIONS['--config_prod']:
        config_class_string = 'abra_api.config.Production'
    else:
        config_class_string = 'abra_api.config.Testing'
    config_obj = get_config(config_class_string)

    return config_obj


def command(name: str = None):
    """Decorator that registers the chosen command/function.

    If a function is decorated with @command but that function name is not a valid "command" according to the docstring,
    a KeyError will be raised, since that's a bug in this script.

    If a user doesn't specify a valid command in their command line arguments, the above docopt(__doc__) line will print
    a short summary and call sys.exit() and stop up there.

    If a user specifies a valid command, but for some reason the developer did not register it, an AttributeError will
    raise, since it is a bug in this script.

    Finally, if a user specifies a valid command and it is registered with @command below, then that command is "chosen"
    by this decorator function, and set as the attribute `chosen`. It is then executed below in
    `if __name__ == '__main__':`.

    Positional arguments:
    func -- the function to decorate
    """

    def function_wrap(func):

        @wraps(func)
        def wrapped():
            return func()

        command_name = name if name else func.__name__

        # Register chosen function.
        if command_name not in OPTIONS:
            raise KeyError('Cannot register {}, not mentioned in docstring/docopt.'.format(command_name))
        if OPTIONS[command_name]:
            command.chosen = func

        return wrapped

    return function_wrap


@command()
def dump():
    config = parse_config()
    setup_logging('dump', logging.DEBUG if config.DEBUG else logging.WARNING)
    data_loader = DataLoader(Path(config.INPUT_DIRECTORY), 'windows-1250')
    for file_info in data_loader.load_files():
        print(file_info)


@command()
def test_report_mail():
    config = parse_config()
    setup_logging('test_report_mail', logging.DEBUG if config.DEBUG else logging.WARNING)
    if not config.REPORT_EMAIL:
        logging.error('No REPORT_EMAIL is configured, exiting...')
        return

    email_sender = EmailSender(config.MAIL_SERVER, config.MAIL_PORT, config.MAIL_USE_SSL, config.MAIL_USE_TLS, config.MAIL_DEFAULT_SENDER)
    email_sender.login(config.MAIL_USERNAME, config.MAIL_PASSWORD)

    report_text_lines = ['Hello world!']

    email_sender.send(config.REPORT_EMAIL, 'AbraApi Sync report test email', '\n'.join(report_text_lines))

    logging.info('Email was sent?')


@command()
def sync():
    config = parse_config()
    setup_logging('sync', logging.DEBUG if config.DEBUG else logging.WARNING)

    data_syncer = DataSyncer(config)
    try:
        data_syncer.sync()
    finally:
        data_syncer.close()


@command()
def watch():
    config = parse_config()
    setup_logging('watch', logging.DEBUG if config.DEBUG else logging.WARNING)

    data_syncer = DataSyncer(config)
    try:
        if OPTIONS['pool']:
            if not OPTIONS['--pool_time'] and not config.POOL_TIME:
                raise ValueError('Please specify --pool_time or POOL_TIME for pool watcher')

            pool_time = OPTIONS['--pool_time'] if OPTIONS['--pool_time'] else config.POOL_TIME

            while True:
                logging.debug('Pooling data...')
                data_syncer.sync()
                time.sleep(int(pool_time))
        elif OPTIONS['inotify']:
            def suffix_filter(event):
                # return True to stop processing of event (to "stop chaining")
                return os.path.splitext(event.name)[1] not in ['.NOV', '.REA']

            class ModHandler(pyinotify.ProcessEvent):
                # evt has useful properties, including pathname
                def process_IN_MODIFY(self, evt):
                    data_syncer.sync(Path(evt.pathname))

                def process_IN_CREATE(self, evt):
                    data_syncer.sync(Path(evt.pathname))

            handler = ModHandler(pevent=suffix_filter)
            wm = pyinotify.WatchManager()
            notifier = pyinotify.Notifier(wm, handler)
            wm.add_watch(config.INPUT_DIRECTORY, pyinotify.IN_MODIFY | pyinotify.IN_CREATE)
            notifier.loop()

    finally:
        data_syncer.close()


def main():
    signal.signal(signal.SIGINT, lambda *_: sys.exit(0))  # Properly handle Control+C
    getattr(command, 'chosen')()  # Execute the function specified by the user.


if __name__ == '__main__':
    main()

