import logging
from abra_api.DataInfo import DataInfo
from abra_api.filters.IFilter import IFilter
from abra_api.enums.DataItemTypeEnum import DataItemTypeEnum
from abra_api.ImapConnector import ImapConnector
from abra_api.ImapConnector import ImapFolder


class DonePdFilter(IFilter):
    imap_connector = None

    required_config = [
        'DONE_PD_IMAP_HOST',  # IMAP host to connect
        'DONE_PD_IMAP_USER',  # IMAP user to connect
        'DONE_PD_IMAP_PASSWORD',  # IMAP password to connect
        'DONE_PD_IMAP_SSL',  # Does IMAP server uses SSL
        'DONE_PD_DIRECTORY_FORMAT',  # Directory that is created in {order_year}/WHOEVER/{order_name}
        'DONE_PD_SEARCH_NEW_PD_DIRECTORY_FORMAT'  # Directory that is created in {order_year}/WHOEVER/{order_name}
    ]

    def _needs_imap_connector(self):
        logging.debug('Connecting to IMAP server...')
        self.imap_connector = ImapConnector(
            self.config.get('DONE_PD_IMAP_HOST'),
            self.config.get('DONE_PD_IMAP_SSL'),
        )
        logging.debug('OK!')
        logging.debug('Logging to IMAP server...')
        self.imap_connector.login(
            self.config.get('DONE_PD_IMAP_USER'),
            self.config.get('DONE_PD_IMAP_PASSWORD')
        )
        logging.debug('OK!')

    def close(self) -> None:
        if self.imap_connector:
            self.imap_connector.close()

    def is_handled(self, data_info: DataInfo) -> bool:
        if data_info.data_item.state == DataItemTypeEnum.DONE and 'PD' in data_info.data_item.order_name_identifier:
            return True

        return False

    def process(self, data_info: DataInfo) -> None:
        self._needs_imap_connector()
        order_year, _ = data_info.data_item.order_name_identifier.split('/', 1)

        format_dict = dict(**data_info.data_item.to_dict())
        format_dict['order_year'] = order_year

        directory_path_parts = self.config.get('DONE_PD_DIRECTORY_FORMAT').split('/')
        search_directory_path_parts = self.config.get('DONE_PD_SEARCH_NEW_PD_DIRECTORY_FORMAT').split('/')

        current_folders = list(self.imap_connector.get_folders())

        # check if folder already exists
        imap_folder = ImapFolder.from_str_list([directory_path_part.format(**format_dict).replace('/', '|') for directory_path_part in directory_path_parts])
        if imap_folder in current_folders:
            logging.info('Folder {} already exists on the server, subscribing...'.format(imap_folder))
            self.imap_connector.subscribe_to_folder(imap_folder)
            return

        # Check if folder is in DONE_PD_SEARCH_NEW_PD_DIRECTORY_FORMAT
        imap_folder_search = ImapFolder.from_str_list([search_directory_path_part.format(**format_dict).replace('/', '|') for search_directory_path_part in search_directory_path_parts])

        if imap_folder_search in current_folders:
            # Rename imap_folder_search to imap_folder
            logging.debug('Renaming folder {} to {}'.format(imap_folder_search, imap_folder))
            self.imap_connector.unsubscribe_from_folder(imap_folder_search)
            self.imap_connector.rename_folder(imap_folder_search, imap_folder)
        else:
            logging.debug('Creating new folder {}'.format(imap_folder))
            self.imap_connector.create_folder(imap_folder)
        self.imap_connector.subscribe_to_folder(imap_folder)

        logging.debug('OK!')

