import logging
from abra_api.DataInfo import DataInfo
from abra_api.filters.IFilter import IFilter
from abra_api.enums.DataItemTypeEnum import DataItemTypeEnum
from abra_api.ImapConnector import ImapConnector
from abra_api.ImapConnector import ImapFolder


class NewPdFilter(IFilter):
    imap_connector = None
    required_config = [
        'NEW_PD_IMAP_HOST',  # IMAP host to connect
        'NEW_PD_IMAP_USER',  # IMAP user to connect
        'NEW_PD_IMAP_PASSWORD',  # IMAP password to connect
        'NEW_PD_IMAP_SSL',  # Does IMAP server uses SSL
        'NEW_PD_DIRECTORY_FORMAT'  # Directory that is created in {ORDER_YEAR}/WHOEVER/{ORDER_NAME}
    ]

    def _needs_imap_connector(self):
        logging.debug('Connecting to IMAP server...')
        self.imap_connector = ImapConnector(
            self.config.get('NEW_PD_IMAP_HOST'),
            self.config.get('NEW_PD_IMAP_SSL'),
        )
        logging.debug('OK!')
        logging.debug('Logging to IMAP server...')
        self.imap_connector.login(
            self.config.get('NEW_PD_IMAP_USER'),
            self.config.get('NEW_PD_IMAP_PASSWORD')
        )
        logging.debug('OK!')

    def close(self) -> None:
        if self.imap_connector:
            self.imap_connector.close()

    def is_handled(self, data_info: DataInfo) -> bool:
        if data_info.data_item.state == DataItemTypeEnum.NEW and 'PD' in data_info.data_item.order_name_identifier:
            return True

        return False

    def process(self, data_info: DataInfo) -> None:
        self._needs_imap_connector()
        order_year, _ = data_info.data_item.order_name_identifier.split('/', 1)

        format_dict = dict(**data_info.data_item.to_dict())
        format_dict['order_year'] = order_year

        directory_path_parts = self.config.get('NEW_PD_DIRECTORY_FORMAT').split('/')

        current_folders = self.imap_connector.get_folders()

        imap_folder = ImapFolder.from_str_list([directory_path_part.format(**format_dict).replace('/', '|') for directory_path_part in directory_path_parts])
        if imap_folder in current_folders:
            logging.info('Folder {} already exists on the server, subscribing...'.format(imap_folder))
            self.imap_connector.subscribe_to_folder(imap_folder)
            return

        logging.debug('Creating folder {}'.format(imap_folder))
        self.imap_connector.create_folder(imap_folder)
        self.imap_connector.subscribe_to_folder(imap_folder)
        logging.debug('OK!')

